local _, BigBrother = ...

local module = BigBrother:NewModule("Alerts")
local L = LibStub("AceLocale-3.0"):GetLocale("BigBrother")

local _G = _G
local bit_band, bit_bor = _G.bit.band, _G.bit.bor

local UNKNOWN = ("(%s)"):format(_G.UNKNOWN)
local soulstoneList = {}

-- Copied from BigWigs_CommonAuras by Rabbit
local combatLogMap = {}
combatLogMap.SPELL_CAST_START = {
	--Feasts/Cauldrons
	[57426] = "Feast", -- Fish Feast
	[57301] = "Feast", -- Great Feast
	[66476] = "Feast", -- Bountiful Feast
	[87643] = "Feast", -- Broiled Dragon Feast
	[87915] = "Feast", -- Goblin Barbecue Feast
	[87644] = "Feast", -- Seafood Magnifique Feast
	[92649] = "Feast", -- Cauldron of Battle
	[92712] = "Feast", -- Big Cauldron of Battle
}
combatLogMap.SPELL_CAST_SUCCESS = {
	--Repair Bots
	[22700] = "Repair",  -- Field Repair Bot 74A
	[44389] = "Repair",  -- Field Repair Bot 110G
	[54711] = "Repair",  -- Scrapbot
	[67826] = "Repair",  -- Jeeves
	--Summoning
	[29893] = "Summon", -- Ritual of Souls
	[698]   = "Summon", -- Ritual of Summoning
	[92827] = "Summon", -- Ritual of Refreshment
	-- Misdirects
	[34477] = "Misdirect", -- Misdirection
	[57934] = "Misdirect", -- Tricks of the Trade
	-- Taunt AoEs
	[1161] = "Taunt",  -- Challenging Shout
	[5209] = "Taunt",  -- Challenging Roar
	[31789] = "Taunt", -- Righteous Defense
}
combatLogMap.SPELL_AURA_APPLIED = {
	-- Taunts
	[355] = "Taunt",   -- Taunt
	[6795] = "Taunt",  -- Growl
	[20736] = "Taunt", -- Distracting Shot
	[51399] = "Taunt", -- Death Grip
	[56222] = "Taunt", -- Dark Command
	[62124] = "Taunt", -- Hand of Reckoning
}
combatLogMap.SPELL_CREATE = {
	-- Portals
	[11419] = "Portal", -- Darnassus
	[32266] = "Portal", -- Exodar
	[11416] = "Portal", -- Ironforge
	[11417] = "Portal", -- Orgrimmar
	[33691] = "Portal", -- Shattrath (Alliance)
	[35717] = "Portal", -- Shattrath (Horde)
	[32267] = "Portal", -- Silvermoon
	[10059] = "Portal", -- Stormwind
	[11420] = "Portal", -- Thunder Bluff
	[11418] = "Portal", -- Undercity
	[49360] = "Portal", -- Theramore
	[49361] = "Portal", -- Stonard
	[53142] = "Portal", -- Dalaran
	[88345] = "Portal", -- Tol Barad (Alliance)
	[88346] = "Portal", -- Tol Barad (Horde)
}
combatLogMap.SPELL_RESSURECT = {
	[20484] = "Ressurect", -- Rebirth
}
combatLogMap.SPELL_INTERRUPT = {
	["*"] = "Interrupt",
}
combatLogMap.SPELL_MISSED = {
	-- Interrupts
	[47528] = "InterruptMiss", -- Knight Mind Freeze
	[47476] = "InterruptMiss", -- Knight Strangulate
	[5211] = "InterruptMiss",  -- Bash
	[34490] = "InterruptMiss", -- Silencing Shot
	[2139] = "InterruptMiss",  -- Counterspell
	[15487] = "InterruptMiss", -- Silence
	[1766] = "InterruptMiss",  -- Kick
	[57994] = "InterruptMiss", -- Wind Shock
	[72] = "InterruptMiss",    -- Shield Bash
	[6552] = "InterruptMiss",  -- Pummel
	-- Taunts
	[355] = "TauntMiss",   -- Taunt
	[6795] = "TauntMiss",  -- Growl
	[20736] = "TauntMiss", -- Distracting Shot
	[51399] = "TauntMiss", -- Death Grip
	[56222] = "TauntMiss", -- Dark Command
	[62124] = "TauntMiss", -- Hand of Reckoning
}
-- Crowd Control (beware the shennanigans)
combatLogMap.SPELL_AURA_REMOVED = {
	[118] = "CrowdControl",   -- Polymorph !
	[28271] = "CrowdControl", -- Polymorph Turtle
	[28272] = "CrowdControl", -- Polymorph Pig
	[61305] = "CrowdControl", -- Polymorph Black Cat
	[61721] = "CrowdControl", -- Polymorph Rabbit
	[61780] = "CrowdControl", -- Polymorph Turkey
	[2094] = "CrowdControl",  -- Blind
	[2637] = "CrowdControl",  -- Hibernate
	[5782] = "CrowdControl",  -- Fear !
	[6770] = "CrowdControl",  -- Sap
	[20066] = "CrowdControl", -- Repentance !
	[51514] = "CrowdControl", -- Hex !
}
combatLogMap.SPELL_AURA_BROKEN_SPELL = { --these three actually trigger the event
	[9484] = "CrowdControl",  -- Shackle Undead !
	[3355] = "CrowdControl",  -- Freezing Trap !
	[6358] = "CrowdControl",  -- Seduction !
}
for k,v in next, combatLogMap.SPELL_AURA_REMOVED do 
	combatLogMap.SPELL_AURA_BROKEN_SPELL[k] = v
end
combatLogMap.SPELL_AURA_BROKEN = combatLogMap.SPELL_AURA_BROKEN_SPELL --don't think broken is ever triggered on melee breaks, but oh well

--specialiness for soulstones. wonder why there isn't a SPELL_RESSURECT event for ss and ankh
combatLogMap.SPELL_AURA_REMOVED[20707] = "Soulstone"

function module:OnInitialize()
	self.db = BigBrother.db:RegisterNamespace("Alerts", {
		profile = {
			crowdControl = true,
			misdirect = true,
			taunt = false,
			interrupt = false,
			combatRes = true,
			portal = true,
			repair = true,
			feast = true,
			summon = true,
			spellLink = true,
			groupOnly = true,
			icons = true,
			fallback = true,
			output = "self",
			outputExtra = "",
		}
	})
end

function module:OnEnable()
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
end

function module:Spam(msg)
	if not msg or msg == "" then return end
	local chatMsg = string.gsub(msg, "|Hicon:%d+:dest|h|TInterface.TargetingFrame.UI%-RaidTargetingIcon_(%d).blp:0|t|h", "{rt%1}")
	chatMsg = string.gsub(chatMsg,"|c%x%x%x%x%x%x%x%x([^%|].-)|r", "%1") --colored text in SendChatMessage can disconnect the user, apparently (spell links are special)
	local output = self.db.profile.output
	local doSelf
	
	if output == "group" then
		if GetNumRaidMembers() > 0 then
			SendChatMessage(chatMsg, "RAID")
		elseif GetNumPartyMembers() > 0 then
			SendChatMessage(chatMsg, "PARTY")
		else
			doSelf = true
		end
	elseif output == "raid" then
		if GetNumRaidMembers() > 0 then
			SendChatMessage(chatMsg, "RAID")
		else
			doSelf = true
		end
	elseif output == "party" then
		if GetNumPartyMembers() > 0 then
			SendChatMessage(chatMsg, "PARTY")
		else
			doSelf = true
		end
	elseif output == "guild" then
		if IsInGuild() and (GetNumPartyMembers() > 0 or GetNumRaidMembers() > 0) then
			SendChatMessage(chatMsg, "GUILD")
		else
			doSelf = true
		end
	elseif output == "officer" then
		if IsInGuild() and (GetNumPartyMembers() > 0 or GetNumRaidMembers() > 0) then
			SendChatMessage(chatMsg, "OFFICER")
		else
			doSelf = true
		end
	elseif output == "channel" then
		local index = GetChannelName(self.db.profile.outputExtra)
		if index > 0 then
			SendChatMessage(chatMsg, "CHANNEL", nil, index)
		else
			doSelf = true
		end
	end
	
	if output == "self" or (self.db.profile.fallback and doSelf) then
		BigBrother:Print(msg)
	end
end


do --module:COMBAT_LOG_EVENT_UNFILTERED
	local FILTER_GROUP = bit_bor(_G.COMBATLOG_OBJECT_AFFILIATION_MINE, _G.COMBATLOG_OBJECT_AFFILIATION_PARTY, _G.COMBATLOG_OBJECT_AFFILIATION_RAID)
	local function checkSrc(flags)
		if module.db.profile.groupOnly then
			return bit_band(flags, FILTER_GROUP) ~= 0
		end
		return true
	end
	
	local function getIconString(flags)
		if module.db.profile.icons then
			local raidTarget = bit_band(flags, _G.COMBATLOG_OBJECT_RAIDTARGET_MASK)
			if raidTarget ~= 0 then
				for i=1,8 do
					local mask = _G.COMBATLOG_OBJECT_RAIDTARGET1 * (2 ^ (i - 1)) --_G["COMBATLOG_OBJECT_RAIDTARGET"..i]
					if raidTarget == mask then
						return _G.TEXT_MODE_A_STRING_DEST_ICON:format(mask, _G["COMBATLOG_ICON_RAIDTARGET"..i])
					end
				end
			end
		end
		return ""
	end
	
	local removedAura
	function module:COMBAT_LOG_EVENT_UNFILTERED(_, ts, event, _, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, _, extraSpellId, extraSpellName)
	
		--create our own SPELL_AURA_BROKEN_SPELL event from SPELL_AURA_REMOVED + SPELL_DAMAGE
		if event == "SPELL_AURA_REMOVED" and combatLogMap["SPELL_AURA_BROKEN_SPELL"][spellId] then
			removedAura = {ts = ts, id = spellId, name = spellName, guid = dstGUID}
			return
		end
		if removedAura then
			--SPELL_AURA_REMOVED and the SPELL_DAMAGE that removed it have the same timestamp, and check the guid just to make sure
			if event == "SPELL_DAMAGE" and removedAura.ts == ts and removedAura.guid == dstGUID then
				event = "SPELL_AURA_BROKEN_SPELL"
				extraSpellId = spellId
				extraSpellName = spellName
				spellId = removedAura.id
				spellName = removedAura.name
			--[===[@debug@
			elseif event == "SWING_DAMAGE" and removedAura.guid == dstGUID then
				--the swing that breaks it looks to be ~400ms (sample size: 2) _before_ SPELL_AURA_REMOVED, bah
				event = "SPELL_AURA_BROKEN"
				extraSpellId = _G.AURA_TYPE_DEBUFF
				extraSpellName = nil
				spellId = removedAura.id
				spellName = removedAura.name
				print(event, spellName, (ts-removedAura.ts))
			--@end-debug@]===]
			end
			removedAura = nil --the events should be back to back
		end
		
		--and now for things that behave nicely
		local func = combatLogMap[event] and (combatLogMap[event][spellId] or combatLogMap[event]["*"]) or nil
		if func and checkSrc(srcFlags) then 
			local srcOutput = ("%s|cff40ff40%s|r"):format(getIconString(srcFlags), string.gsub(srcName or UNKNOWN, "(%a)%-(.*)", "%1"))
			local dstOutput = ("%s|cffff4040%s|r"):format(getIconString(dstFlags), string.gsub(dstName or UNKNOWN, "(%a)%-(.*)", "%1"))
			local spellOutput = self.db.profile.spellLink and GetSpellLink(spellId) or spellName
			 --was for cc, not sure what _BROKEN would trigger from on friendlies. mc -> cc -> mc fades (still cc'd) -> mob breaks cc? or pvp i suppose
			--local hostile = bit_band(dstFlags, _G.COMBATLOG_OBJECT_REACTION_HOSTILE) ~= 0
			
			self[func](self, srcOutput, dstOutput, spellOutput, extraSpellId, extraSpellName)
		end
		
		if event == "UNIT_DIED" and soulstoneList[dstName] then
			soulstoneList[dstName] = GetTime() + 60 --you have 6min to use it (auto-release), but if you hold onto it for more than a minute...prolly don't care
		end
		
	end
end


---------------------------------------
-- Spell handlers

do --CrowdControl
	local spamCount = 0
	local spamLastTime = 0
	function module:CrowdControl(srcOutput, dstOutput, spellOutput, extraSpell, extraSpellName)
		local output
		if extraSpellName then --SPELL_AURA_BROKEN_SPELL
			extraSpellName = self.db.profile.spellLink and GetSpellLink(extraSpell) or extraSpellName
			output = L["%s on %s removed by %s's %s"]:format(spellOutput, dstOutput, srcOutput, extraSpellName)
		else --SPELL_AURA_BROKEN / SPELL_AURA_REMOVED
			output = L["%s on %s removed by %s"]:format(spellOutput, dstOutput, srcOutput)
		end
		
		--throttle
		local now = GetTime()
		if now - spamLastTime > 3 then
			spamCount = 0
		end
		spamCount = spamCount + 1
		if spamCount > 3 then
			output = nil
		end
		spamLastTime = now
		
		self:Spam(output)
	end
end

function module:Misdirect(srcOutput, dstOutput, spellOutput)
	if not self.db.profile.misdirect then return end
	self:Spam(L["%s cast %s on %s"]:format(srcOutput, spellOutput, dstOutput))
end

function module:Taunt(srcOutput, dstOutput, spellOutput)
	if not self.db.profile.taunt then return end
	if dstOutput then --SPELL_AURA_APPLIED
		self:Spam(L["%s cast %s on %s"]:format(srcOutput, spellOutput, dstOutput))
	else --SPELL_CAST_SUCCESS
		self:Spam(L["%s cast %s"]:format(srcOutput, spellOutput))
	end
end

function module:TauntMiss(srcOutput, dstOutput, spellOutput, extraSpell)
	if not self.db.profile.taunt then return end
	self:Spam(L["%s cast %s on %s (%s)"]:format(srcOutput, spellOutput, dstOutput, _G["ACTION_SPELL_MISSED_"..extraSpell]))
end

function module:Interrupt(srcOutput, dstOutput, spellOutput, extraSpell, extraSpellName)
	if not self.db.profile.interrupt then return end
	extraSpellName = self.db.profile.spellLink and GetSpellLink(extraSpell) or extraSpellName
	self:Spam(L["%s interrupted %s's %s with %s"]:format(srcOutput, dstOutput, extraSpellName, spellOutput))
end

function module:InterruptMiss(srcOutput, dstOutput, spellOutput, extraSpell)
	if not self.db.profile.interrupt then return end
	self:Spam(L["%s cast %s on %s (%s)"]:format(srcOutput, spellOutput, dstOutput, _G["ACTION_SPELL_MISSED_"..extraSpell]))
end

function module:Ressurect(srcOutput, dstOutput, spellOutput)
	if not self.db.profile.combatRes then return end
	self:Spam(L["%s cast %s on %s"]:format(srcOutput, spellOutput, dstOutput))
end

function module:Portal(srcOutput, _, spellOutput)
	if not self.db.profile.portal then return end
	self:Spam(L["%s used %s"]:format(srcOutput, spellOutput, dstOutput))
end

function module:Repair(srcOutput, _, spellOutput)
	if not self.db.profile.repair then return end
	self:Spam(L["%s used %s"]:format(srcOutput, spellOutput))
end

function module:Feast(srcOutput, _, spellOutput)
	if not self.db.profile.feast then return end
	self:Spam(L["%s used %s"]:format(srcOutput, spellOutput))
end

function module:Summon(srcOutput, _, spellOutput)
	if not self.db.profile.summon then return end
	self:Spam(L["%s is casting %s"]:format(srcOutput, spellOutput))
end

do --Soulstone
	local spiritOfRedemption = GetSpellInfo(27827) --still fires UNIT_DIED twice. works out, i suppose
	
	local f = CreateFrame("Frame")
	local total = 0
	local function soulstoneCheck(self, elapsed)
		total = elapsed + total
		if total > 1 then
			total = 0
			
			local now = GetTime()
			for name, expires in next, soulstoneList do
				--print(name, string.format("%.01f", (expires - now)), "|", (now > expires), UnitIsGhost(name), not (UnitIsDead(name) or UnitIsFeignDeath(name)))
				if now > expires or UnitIsGhost(name) then --expired or released
					soulstoneList[name] = nil
				elseif not UnitIsDead(name) and not UnitIsFeignDeath(name) and not UnitBuff(name, spiritOfRedemption) then
					soulstoneList[name] = nil
					local spellOutput = module.db.profile.spellLink and GetSpellLink(6203) or GetSpellInfo(6203) --20707 Soulstone Resurrection
					module:Spam(L["%s used %s"]:format(("|cff40ff40%s|r"):format(name), spellOutput))
				end
			end
			
			if not next(soulstoneList) then
				self:SetScript("OnUpdate", nil)
			end
		end
	end

	function module:Soulstone(_, dstOutput)
		if not self.db.profile.combatRes then return end
		
		local name = string.match(dstOutput, "|cffff4040(.-)|r")
		if name and not UnitName(name) then return end
		
		--print("Soulstone", name)
		soulstoneList[name] = GetTime() + .5 --timeout for UNIT_DIED
		f:SetScript("OnUpdate", soulstoneCheck)
	end
end


---------------------------------------
-- Options

function module:AddOptions(options)
	options.args.alerts = {
		name = L["Alerts"],
		--desc = L["Alerts"],
		type = "group",
		get = function(info) return self.db.profile[info[#info]] end,
		set = function(info, value) self.db.profile[info[#info]] = value end,
		order = 20,
		args = {
			report = {
				name = L["Combat Alerts"],
				desc = L["Set what is reported to chat."],
				type = "group",
				inline = true,
				order = 10,
				args = {
					crowdControl = {
						name  = L["Crowd Control"],
						desc = L["Report when a player breaks a crowd control effect."],
						type = "toggle",
						order = 5,
					},
					misdirect = {
						name  = L["Misdirect"],
						desc = L["Report who gains Misdirection."],
						type = "toggle",
						order = 10,
					},
					taunt = {
						name  = L["Taunt"],
						desc = L["Report taunts."],
						type = "toggle",
						order = 15,
					},
					interrupt = {
						name  = L["Interrupts"],
						desc = L["Report interrupts."],
						type = "toggle",
						order = 16,
					},
					combatRes = {
						name = L["Combat Resurrection"],
						desc = L["Report battle resurrections (Druid Rebirth)."],
						type = "toggle",
						order = 17,
					},
				},
			}, --report
			
			report2 = {
				name = L["Noncombat Alerts"],
				desc = L["Set what is reported to chat."],
				type = "group",
				inline = true,
				order = 11,
				args = {
					feast = {
						name = L["Consumables"],
						desc = L["Report when a player uses a feast or a cauldron."],
						type = "toggle",
						order = 5,
					},
					repair = {
						name = L["Repair Bots"],
						desc = L["Report when a player uses a repair bot."],
						type = "toggle",
						order = 10,
					},
					portal = {
						name = L["Portals"],
						desc = L["Report when a Mage opens a portal."],
						type = "toggle",
						order = 15,
					},
					summon = {
						name = L["Rituals"],
						desc = L["Report when a player needs assistance summoning an object."],
						type = "toggle",
						order = 16,
					},
				},
			}, --report2
			
			output = {
				name = L["Output"],
				desc = L["Set where the alert output is sent."],
				type = "group",
				inline = true,
				get = function(info)
					local key = info[#info]
					return self.db.profile.output == key
				end,
				set = function(info, value)
					if value then
						local key = info[#info]
						self.db.profile.output = key
					end
				end,
				order = 20,
				args = {
					self = {
						name = L["Self"],
						desc = L["Print to your chat window."],
						type = "toggle",
						order = 5,
					},
					group = {
						name = L["Group"],
						desc = L["Report to party or raid chat."],
						type = "toggle",
						order = 10,
					},
					party = {
						name = L["Party"],
						desc = L["Report to party chat."],
						type = "toggle",
						order = 11,
					},
					raid = {
						name = L["Raid"],
						desc = L["Report to raid chat."],
						type = "toggle",
						order = 12,
					},
					guild = {
						name = L["Guild"],
						desc = L["Report to guild chat."],
						type = "toggle",
						order = 15,
					},
					officer = {
						name = L["Officer"],
						desc = L["Report to officer chat."],
						type = "toggle",
						order = 20,
					},
					channel = {
						name = L["Channel"],
						desc = L["Report to a custom channel."],
						type = "toggle",
						order = 25,
					},
					extra = {
						name = L["Channel"],
						type = "select",
						values = BigBrother.GetChatChannels,
						get = function() return self.db.profile.outputExtra end,
						set = function(info, value) self.db.profile.outputExtra = value end,
						hidden = function() return self.db.profile.output ~= "channel" end,
						order = 101,
					},
				},
			},
			
			spellLink = {
				name = L["Use Spell Links"],
				desc = L["Display spell names as clickable spell links."],
				type = "toggle",
				order = 40,
			},
			groupOnly = {
				name = L["Group Only"],
				desc = L["Only report events from players in your party or raid group."],
				type = "toggle",
				order = 41,
			},
			icons = {
				name = L["Raid Target Icons"],
				desc = L["Show associated raid target icons as part of the alert."],
				type = "toggle",
				order = 42,
			},
			fallback = {
				name = L["Fallback Output to Self"],
				desc = L["Print to the chat window if unable to send to the specified output."],
				type = "toggle",
				order = 44,
			},
			
		},
	}
end
